#ifndef PARAMETRIC_MESH_H
#define PARAMETRIC_MESH_H

#include "mesh.h"
#include <cmath>

class MeshGenerator
{
public:
    virtual glm::vec3 evalUV(float u, float v) = 0;
    
    /**
     * @brief generateParametricMesh creates a mesh from a rectangular grid of size width:height
     */
    Mesh* generateParametricMesh(Material* mat, int width = 10, int height = 10, float size = 1, bool alternate = true);
    
    /**
     * @brief generateGeodesicMesh creates a closed mesh from an icosahedron subdivided n times
     */
    Mesh* generateGeodesicMesh(Material* mat, int n = 0, float size = 1);
private:
    float m_size;
    Mesh *m_mesh;

    // geodesic methods
    class Edge{
    public:
        int b;
        int vertex;
        Edge* next;
        Edge(int myB = -1, int myVertex = -1) :
            b(myB),
            vertex(myVertex),
            next(NULL)
        {}
        ~Edge(){ if(next != NULL) delete(next); }
    };

    Edge* edges;

    int getEdge(int a, int b);
    void createVertex(float u, float v);
    void subdivide();
    
    // parametric methods
    int getVertexId(int i, int j, int height);
};

// here are some basic implementations of parametric mesh generators

/**
 * @brief The GridGenerator class is the simplest parametric mesh, a uniform mapped grid
 */
class GridGenerator : public MeshGenerator
{
public:
    virtual glm::vec3 evalUV(float u, float v)
        {return glm::vec3(u*2-1, 0, v*2-1);}
};

/**
 * @brief The CylinderGenerator class creates a vertical open cylinder
 */
class CylinderGenerator : public MeshGenerator
{
public:
    virtual glm::vec3 evalUV(float u, float v)
        {return glm::vec3(cos(u*6.2832f), v*2-1, sin(u*6.2832f));}
};

/**
 * @brief The ConeGenerator class creates an open cone pointing up
 */
class ConeGenerator : public MeshGenerator
{
public:
    virtual glm::vec3 evalUV(float u, float v)
        {return glm::vec3(cos(-u*6.2832f)*v, 1 - v*2, sin(-u*6.2832f)*v);}
};

/**
 * @brief The SphereGenerator class creates a sphere uv-mapped like a planisphere
 * it is recommended to use a grid where the width value is twice the value of height
 */
class SphereGenerator : public MeshGenerator
{
public:
    virtual glm::vec3 evalUV(float u, float v)
    {
        float r = sin(v*3.1416f);
        return glm::vec3(cos(-u*6.2832f)*r, cos(v*3.1416f), sin(-u*6.2832f)*r);
    }
};

#endif // PARAMETRIC_MESH_H
